#include <HALOptions.h>
#include <SerialChipRegistry.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "CircleBuf.h"
#include "BluntServer.h"
#include "HCI.h"
#include "L2CAP.h"
#include "RFCOMM.h"
#include "SDP.h"

#include "Logger.h"

#include <NewtonGestalt.h>

typedef long Ref;

const Ref	NILREF = 0x02;
const Ref	TRUEREF = 0x1A;
const Ref	FALSEREF = NILREF;	

class TObjectIterator;
class RefVar;
typedef const RefVar& RefArg;

typedef Ref (*MapSlotsFunction)(RefArg tag, RefArg value, ULong anything);

extern Ref 	MakeInt(long i);
extern Ref 	MakeChar(unsigned char c);
extern Ref	MakeBoolean(int val);

extern Boolean IsInt(RefArg r);	
extern Boolean IsChar(RefArg r);	
extern Boolean IsPtr(RefArg r);		
extern Boolean IsMagicPtr(RefArg r);    
extern Boolean IsRealPtr(RefArg r);   
					
extern long RefToInt(RefArg r);			
extern UniChar RefToUniChar(RefArg r);		

#define kReceiveBufReg 3
#define kTransmitBufReg 3
#define kInterruptEnbReg 2
#define kInterruptIDReg 1
#define kLineControlReg 0
#define kModemControlReg 7
#define kLineStatusReg 6
#define kModemStatusReg 5
#define kScratchReg 4

#define DEBUGSTR
// #define DEBUGSTR DebugStr
// #define DEBUGSTR printf

#define EVT_BLUNT_CHECK_INPUT 0

extern "C" void EnterFIQAtomic (void);
extern "C" void ExitFIQAtomic (void);
extern "C" void _EnterFIQAtomic (void);
extern "C" void _ExitFIQAtomicFast (void);
extern TUPort* GetNewtTaskPort (void);
extern "C" long LockStack(TULockStack* lockRef, ULong additionalSpace);	// Lock the entire stack plus additionalSpace # of bytes
extern "C" long UnlockStack(TULockStack* lockRef);
extern "C" long LockHeapRange(VAddr start, VAddr end, Boolean wire);	// Lock range from start up to but not including end.  Wire will prevent v->p mappings from changing.
extern "C" long UnlockHeapRange(VAddr start, VAddr end);
extern "C" _sys_write (int fd, char* data, int len);

// ================================================================================
//  Interrupt Handlers
// ================================================================================

void TxBEmptyIntHandler (void* server)
{
	DEBUGSTR ("TxBEmptyIntHandler\n");
	((BluntServer*)server)->TxBEmptyIntHandler ();
}

void RxCAvailIntHandler (void* server)
{
	DEBUGSTR ("RxCAvailIntHandler\n");
	 ((BluntServer*)server)->RxCAvailIntHandler ();
}

void RxCSpecialIntHandler (void* server)
{
	DEBUGSTR ("RxCSpecialIntHandler\n");
	((BluntServer*)server)->RxCSpecialIntHandler ();
}

void RxDmaIntHandler (void* server, RxErrorStatus status)
{
	DEBUGSTR ("RxDmaIntHandler\n");
	 ((BluntServer*)server)->RxDMAIntHandler (status);
}

void TxDmaIntHandler (void* server)
{
	DEBUGSTR ("TxDmaIntHandler\n");
	 ((BluntServer*)server)->TxDMAIntHandler ();
}

void ExtStsIntHandler (void* server)
{
	DEBUGSTR ("ExtStsIntHandler\n");
//	 ((BluntServer*)server)->ExtStsIntHandler ();
}

void __EndOfInterruptHandlers ()
{
}

#pragma mark -

// ================================================================================
//  TExtendedCircleBuf Functions
// ================================================================================

UByte TExtendedCircleBuf::PeekByte (ULong offset)
{
	return fBuffer[(fStart + offset) % fSize];
}

#pragma mark -

// ================================================================================
//  Driver Specific Functions
// ================================================================================

class TGPIOInterface
{
public:
	NewtonErr ReadGPIOData (UByte, ULong *);
};

class TBIOInterface
{
public:
	NewtonErr ReadDIOPins (UByte pin, ULong *data);
	NewtonErr WriteDIODir (UByte pin, UByte dir, UByte *data);

	NewtonErr WriteDIOPins (UByte pin, UByte value, UByte *data);
};

class TVoyagerPlatform
{
public:
	char filler_0000[0x0010];
	
	TGPIOInterface *fGPIOInterface;
	TBIOInterface *fBIOInterface;
	
	char filler_0014[0x00f0];
};

extern TVoyagerPlatform *GetPlatformDriver (void);

static UByte SetChannel3Selector (int value)
{
	TVoyagerPlatform *p;
	UByte data;
	
	p = GetPlatformDriver ();
	p->fBIOInterface->WriteDIOPins (0x22, value, &data);
	return data;
}

void BluntServer::DriverReset (void)
{
	TSerialChip16450 *chip;
	Byte prescale;
	
	HLOG (1, "BluntServer::DriverReset (%d)\n", fDriver);
	chip = (TSerialChip16450 *) fChip;
	HLOG (1, "  Features: %08x\n", fChip->GetFeatures ());
	fDMA = false;
	if (fDriver == DRIVER_BT2000) {
		// Enable Enhanced Mode to get larger FIFO buffers
		chip->WriteSerReg (kLineControlReg, 0x83);
		chip->WriteSerReg (kInterruptIDReg, 0x21);
		chip->WriteSerReg (kLineControlReg, 0x03);
	} else if (fDriver == DRIVER_BT2000E ||
		fDriver == DRIVER_950_7MHZ ||
		fDriver == DRIVER_950_14MHZ ||
		fDriver == DRIVER_950_18MHZ ||
		fDriver == DRIVER_950_32MHZ) {
		// Enable '950 Mode
		chip->WriteSerReg (kLineControlReg, 0xbf);
		chip->WriteSerReg (kInterruptIDReg, 0x10);
		chip->WriteSerReg (kLineControlReg, 0x03);

		// Enable Enhanced Mode to get larger FIFO buffers
		chip->WriteSerReg (kLineControlReg, 0x83);
		chip->WriteSerReg (kInterruptIDReg, 0x21);
		chip->WriteSerReg (kLineControlReg, 0x03);

		// Enable the prescaler
		chip->WriteSerReg (kModemControlReg, 0x80);
		chip->WriteSerReg (kScratchReg, 0x01);
		chip->WriteSerReg (kLineStatusReg, 0x20);

		// Set the prescale value
		if (fDriver == DRIVER_BT2000E || fDriver == DRIVER_950_7MHZ) {
			prescale = 0x20;
		} else if (fDriver == DRIVER_950_14MHZ) {
			prescale = 0x40;
		} else if (fDriver == DRIVER_950_18MHZ) {
			prescale = 0x50;
		} else if (fDriver == DRIVER_950_32MHZ) {
			prescale = 0x8b;
		}
		chip->WriteSerReg (kScratchReg, 0x01);
		chip->WriteSerReg (kLineStatusReg, prescale);
	} else if (fDriver == DRIVER_TAIYO_YUDEN) {
	/*
		SetChannel3Selector (1);
		Wait (100);
		SetChannel3Selector (0);
	*/
		fDMA = true;
    }
	Wait (100);
}

void BluntServer::DriverSendDelay (void)
{
	switch (fDriver) {
		case DRIVER_TAIYO_YUDEN:
			Wait (1);
			break;
		default:
			break;
	}
}

#pragma mark -

// ================================================================================
//  Generic Protocol Handler Specific Functions
// ================================================================================

Handler::Handler ()
{
	fNumHandlers = 0;
	fServer = NULL;
	fParentHandler = NULL;
    fLogLevel = 0;
}

Handler::~Handler ()
{
}

void Handler::AddHandler (Handler *handler, BluntServer *server)
{
	if (fNumHandlers < MAX_HANDLERS) {
		fHandlers[fNumHandlers] = handler;
		fHandlers[fNumHandlers]->fServer = server;
		handler->fParentHandler = this;
		fNumHandlers++;
	} else {
		HLOG (0, "**** %08x number of handlers exceeded ****\n", this);
	}
}

void Handler::RemoveHandler (Handler *handler)
{
	int i;
	
	for (i = 0; i < fNumHandlers; i++) {
		if (fHandlers[i] == handler) {
			memmove (&fHandlers[i], &fHandlers[i + 1],
				sizeof (Handler *) + fNumHandlers - 1);
			fNumHandlers--;
			break;
		}
	}
}

void Handler::Print (int depth)
{
	int i, j;
	char buffer[80];
	
	for (j = 0; j < depth; j++) buffer[j] = ' ';
	buffer[j] = 0;
	if (fHandlerType == H_HCI) HLOG (0, "%sHCI (%04x)", buffer, ((HCI *) this)->fConnectionHandle);
	else if (fHandlerType == H_L2CAP) HLOG (0, "%sL2CAP", buffer);
	else if (fHandlerType == H_SDP) HLOG (0, "%sSDP", buffer);
	else if (fHandlerType == H_RFCOMM) HLOG (0, "%sRFCOMM", buffer);
	HLOG (0, " (%08x)\n", this);
	for (i = 0; i < fNumHandlers; i++) {
		fHandlers[i]->Print (depth + 4);
	}
}

void Handler::SetLogLevel (Long level)
{
	int i;
	
	HLOG (0, "Handler::SetLogLevel %08x %d\n", this, level);
	fLogLevel = level;
	for (i = 0; i < fNumHandlers; i++) {
		fHandlers[i]->SetLogLevel (level);
	}
}

void Handler::Log (int logLevel, char *format, ...)
{
    va_list args;
    char buffer[128];
    
    if (fLogLevel >= logLevel) {
        va_start (args, format);
        vsprintf (buffer, format, args);
        va_end (args);
        _sys_write (1, buffer, strlen (buffer));
    }
}

#pragma mark -

// ================================================================================
//  Server Functions
// ================================================================================

NewtonErr BluntServer::Initialize (ULong location, ULong driver, ULong speed)
{
	PSerialChipRegistry* reg;
	SerialChipID id;
	NewtonErr r = noErr;

    fServer = this;
	fLogLevel = 0;
	fDebug = -1;
	fBytesRead = 0;
	fBufferOutput = false;
	HLOG (1, "BluntServer::Initialize\n");
	fLocation = location;
	fDriver = driver;
	fSpeed = speed;
	fChip = NULL;
	fSemaphore = 0;
	reg = GetSerialChipRegistry ();
	id = reg->FindByLocation (fLocation);
	fChip = reg->GetChipPtr (id);
	HLOG (1, "Driver: %d, Speed: %d Chip: %d (%08x)\n", fDriver, fSpeed, id, fChip);
	if (fChip == NULL) {
		HLOG (1, "No chip found at this location!\n");
		r = kErrNoHW;
	} else {
		* (Long *) ((Byte*) (fChip) + 0x10) = 0;
	}
	return r;
}

ULong BluntServer::GetSizeOf ()
{
	return sizeof (BluntServer);
}

void BluntServer::TxBEmptyIntHandler (void)
{
	ULong n, size;
	int i;
	UByte c;
	
	fSemaphore = 1;
    EnterFIQAtomic ();
    if (fDMA) {
        if (fTxDMABuffer->BufferCount () > 0) {
            fChip->TxDMAControl (kDMAStart);
        } else {
            fChip->ResetTxBEmpty ();
        }
    } else {
        fChip->SetTxDTransceiverEnable (true);
        if (fOutputHead == fOutputTail) {
            fChip->ResetTxBEmpty ();
        } else while (fChip->TxBufEmpty () && fOutputHead != fOutputTail) {
            c = fOutputBuffer[fOutputTail];
            fOutputTail = (fOutputTail + 1) % sizeof (fOutputBuffer);
            fChip->PutByte (c);
        }
    }
    ExitFIQAtomic ();
	fSemaphore = 0;
}

void BluntServer::ExtStsIntHandler (void)
{
}

void BluntServer::RxCAvailIntHandler (void)
{
	UByte c;
	ULong r;
	ULong n;
	
	r = 0;
	fRxErrorStatus = 0;
	while (fChip->RxBufFull () && r == 0) {
		n = 1;
		c = fChip->GetByte ();
		r = fRxDMABuffer->CopyIn (&c, &n, false, 0);
		if (r != 0) fRxErrorStatus = -1;
	}
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}

void BluntServer::RxCSpecialIntHandler (void)
{
	UByte c;
	
	fRxErrorStatus = fChip->GetRxErrorStatus ();
	c = fChip->GetByte ();
}

void BluntServer::RxDMAIntHandler (RxErrorStatus status)
{
	ULong n;
	
	fSemaphore = 1;
	fRxErrorStatus = status;
	fChip->RxDMAControl (kDMANotifyOnNext | kDMASync);
	fSemaphore = 0;
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}
/*
void BluntServer::RxCAvailIntHandler (void)
{
	UByte c;
	
	while (fChip->RxBufFull ()) {
		fRxStatus = fChip->GetRxErrorStatus ();
		c = fChip->GetByte ();
		fInputBuffer[fInputHead] = c;
		fInputHead = (fInputHead + 1) % MAX_INPUT;
		fInputCurrent++;
	}
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}

void BluntServer::RxDMAIntHandler (RxErrorStatus status)
{
	fRxErrorStatus = status;
	fChip->RxDMAControl (kDMANotifyOnNext | kDMASync);
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}
*/

void BluntServer::TxDMAIntHandler ()
{
}

long BluntServer::TaskConstructor ()
{
	TUNameServer nameServer;
	TCMOSerialIOParms parms;
	SCCChannelInts handlers;
	NewtonErr r;
	ULong n;
	HCI* hci;
	TULockStack lock;
	
	HLOG (1, "BluntServer::TaskConstructor\n");
//	memset (fInputBuffer, 0, sizeof (fInputBuffer));
	fOutputHead = 0;
	fOutputTail = 0;
	fInputCurrent = 0;
	fInputHead = 0;
	fInputTail = 0;
    fDMA = true;

	LockHeapRange((VAddr) &::TxBEmptyIntHandler, (VAddr) &::__EndOfInterruptHandlers, true);

	fRxDMABuffer = new TExtendedCircleBuf ();
	fRxDMABuffer->Allocate (MAX_INPUT, 8, 2, 0);
	fTxDMABuffer = new TExtendedCircleBuf ();
	fTxDMABuffer->Allocate (MAX_OUTPUT, 8, 2, 0);
	fChip->SetSerialMode (kSerModeAsync);

	HLOG (1, "Server: %08x Debug: %08x Buffer: %08x\n", this, &fDebug, fRxDMABuffer);

	if (!fChip->PowerIsOn ()) fChip->PowerOn ();
	HLOG (1, "Power: %d\n", fChip->PowerIsOn ());

	DriverReset ();
	fChip->SetInterruptEnable (false);

	LockStack (&lock, 64);
	_EnterFIQAtomic ();

	handlers.TxBEmptyIntHandler = ::TxBEmptyIntHandler;
	handlers.ExtStsIntHandler = ::ExtStsIntHandler;
	handlers.RxCAvailIntHandler = ::RxCAvailIntHandler;
	handlers.RxCSpecialIntHandler = ::RxCSpecialIntHandler;
	
	r = fChip->InstallChipHandler (this, &handlers);
	if (fDMA) {
		fRxDMABuffer->DMABufInfo (&n, NULL, NULL, NULL);
		fTxDMABuffer->DMABufInfo (&n, NULL, NULL, NULL);
		fChip->InitRxDMA (fRxDMABuffer, 0, RxDmaIntHandler);
		fChip->InitTxDMA (fTxDMABuffer, TxDmaIntHandler);
        fChip->RxDMAControl (kDMANotifyOnNext | kDMAStart);
    }
	fChip->SetIntSourceEnable (kSerIntSrcTxBufEmpty, true);
	parms.fStopBits = k1StopBits;
	parms.fParity = kNoParity;
	parms.fDataBits = k8DataBits;
	parms.fSpeed = fSpeed;
	fChip->SetIOParms (&parms);
    fChip->SetSpeed (fSpeed);
	fChip->Reconfigure ();
	fChip->SetInterruptEnable (true);
	fChip->SetTxDTransceiverEnable (true);
	_ExitFIQAtomicFast ();
	UnlockStack (&lock);

	HLOG (1, "Chip configured, DMA: %d\n", fDMA);
	
	fNumHandlers = 0;
	fHandlers = NULL;
	hci = new HCI ();
	hci->fConnectionHandle = -1;
	AddHandler ((Handler *) hci);
	
	fPort.Init ();
	fIntMessage.Init (false);
	HLOG (1, "Server port: %08x\n", fPort.fId);
	nameServer.RegisterName ("BluntServer", "TUPort", fPort.fId, 0);
	
	fOutstandingPackets = 0;
	
	return noErr;
}

void BluntServer::TaskDestructor ()
{
	TUNameServer nameServer;
	int i;
	
	HLOG (1, "BluntServer::TaskDestructor\n  Got %d bytes\n  ", fRxDMABuffer->BufferCount ());
	fChip->SetInterruptEnable (false);
	fChip->RemoveChipHandler (this);
	fChip->SetTxDTransceiverEnable (false);
	fChip->PowerOff ();
	for (i = 0; i < fRxDMABuffer->BufferCount() && i < 256; i++) {
		HLOG (1, "%02x ", fRxDMABuffer->PeekByte (i));
		if ((i + 1) % 16 == 0) HLOG (1, "\n  ");
	}
	* (Long *) ((Byte*) (fChip) + 0x10) = 0;
	nameServer.UnRegisterName ("BluntServer", "TUPort");
	UnlockHeapRange((VAddr) &::TxBEmptyIntHandler, (VAddr) &::__EndOfInterruptHandlers);
	HLOG (1, "Done\n");
    Sleep (100 * kMilliseconds);
}

void BluntServer::TaskMain ()
{
	ULong type = 0;
	Boolean end = false;
	TUMsgToken token;
	ULong n;
	
	HLOG (1, "BluntServer::TaskMain\n");
	while (!end) {
		n = MAX_MESSAGE;
		fPort.Receive (&n, fMessage, MAX_MESSAGE, &token, &type);
		HLOG (1, "--> Got the message %d (%08x): %d bytes (%08x) \n", type, fMessage, n, &token);
		switch (type) {
			case M_DATA:
				HandleData ();
				break;
			case M_COMMAND:
				HandleCommand ((BluntCommand *) fMessage);
				break;
			case M_TIMER:
				HandleTimer ((BluntTimerEvent*) fMessage);
				break;
			case M_QUIT:
				end = true;
				break;
			default:
				break;
		}
	}
}

void BluntServer::StartOutput ()
{
	UByte c;
    ULong n;
    int i;
	
	fBufferOutput = false;
	if (fDMA) {
        HLOG (1, "BluntServer::StartOutput (%d)\n", fTxDMABuffer->BufferCount ());
		fChip->SetTxDTransceiverEnable (false);
        fChip->RxDMAControl (kDMANotifyOnNext | kDMAStart);
		fChip->SetTxDTransceiverEnable (true);
        HLOG (1, " Sent %d bytes (%d left)\n", i, fTxDMABuffer->BufferCount ());
	} else {
        HLOG (1, "BluntServer::StartOutput (%d %d)\n", fOutputHead, fOutputTail);
		fChip->SetTxDTransceiverEnable (false);
		i = 0;
		while (fChip->TxBufEmpty () && fOutputHead != fOutputTail) {
            i++;
            c = fOutputBuffer[fOutputTail];
            fOutputTail = (fOutputTail + 1) % sizeof (fOutputBuffer);
			fChip->PutByte (c);
/*            
			if (!fChip->TxBufEmpty ()) {
				HLOG (1, "*** Send buffer full after %d bytes, %d %d remaining\n", i, fOutputHead, fOutputTail);
			}
*/
		}
		fChip->SetTxDTransceiverEnable (true);

        HLOG (1, " Sent %d bytes (%d %d left)\n", i, fOutputHead, fOutputTail);
	}
}

void BluntServer::Output (UByte* data, ULong size, Boolean packetStart)
{
	ULong n, processedBytes;
	int i;
	
	HLOG (1, "BluntServer::Output: %08x %d\n", data, size);
	
	i = 0;
	while (i < size && i < 96) {
		HLOG (3, "%02x ", data[i]);
		i++;
		if ((i + 1) % 16 == 0) HLOG (3, "\n");
	}
	HLOG (3, "\n");
	
	if (packetStart) {
		fOutstandingPackets++;
	}
	
    if (fDMA) {
        n = size;
        fTxDMABuffer->CopyIn (data, &n, false, 0);
        processedBytes = size - n;
        HLOG (1, " Buffered %d of %d bytes (%d free)\n", processedBytes, size, fTxDMABuffer->BufferSpace ());
    } else {
        for (i = 0; i < size; i++) {
            fOutputBuffer[fOutputHead] = data[i];
            fOutputHead = (fOutputHead + 1) % sizeof (fOutputBuffer);
            if (fOutputTail == fOutputHead) break;
        }
        if (fOutputTail == fOutputHead) {
            HLOG (0, "*** Output buffer overflow (%d of %d bytes written)\n", i, size);
        }
    }
	if (!fBufferOutput) {
		StartOutput ();
	}
}

/*
void BluntServer::CopyRawData ()
{
	ULong n = 0;

	HLOG (2, "BluntServer::CopyRawData\n");
	n = sizeof (fInputBuffer) - fInputCurrent;
	if (sizeof (fInputBuffer) > fInputCurrent) {
		HLOG (2, "  (1) %d %d %d %d\n", fRxDMABuffer->fSize, fRxDMABuffer->fStart, fRxDMABuffer->fNext, fInputCurrent);
		fRxDMABuffer->CopyOut (fInputBuffer + fInputCurrent, &n, &fRxStatus);
		HLOG (2, "  (2) %d %d %d %d\n", fRxDMABuffer->fSize, fRxDMABuffer->fStart, fRxDMABuffer->fNext, n);
		fChip->RxDMAControl (kDMANotifyOnNext | kDMAStart);
		fLastRead = sizeof (fInputBuffer) - fInputCurrent - n;
		if (fInputCurrent + fLastRead < sizeof (fInputBuffer)) {
			fInputCurrent += fLastRead;
			if (fLastRead > 0) {
				HandleData ();
			}
		} else {
			HLOG (0, "*** Overflow 2: %d %d\n", fInputCurrent, n);
		}
	} else {
		HLOG (0, "*** Overflow 1: %d\n", fInputCurrent);
	}
}
*/

void BluntServer::HandleData ()
{
	int i;
	Long lastAmount;
	ULong n;
	
//	HLOG (2, "BluntServer::HandleData %08x %d %d\n", fInputBuffer, fInputCurrent, fRxDMABuffer->BufferCount ());
	HLOG (1, "BluntServer::HandleData %08x %d\n", fRxDMABuffer, fRxDMABuffer->BufferCount ());
	if (fRxErrorStatus != 0) HLOG (0, "*** Circle buffer overflow\n");
	lastAmount = -1;
	do {
		lastAmount = fRxDMABuffer->BufferCount ();
		for (i = 0; fRxDMABuffer->BufferCount () > 0 && i < fNumHandlers; i++) {
			HLOG (2, "BluntServer Handler: %d (%08x), data: %d bytes\n", i, fHandlers[i], fRxDMABuffer->BufferCount ());
			if (fHandlers[i]->fConnectionState != CONN_REMOVE) {
				((HCI *) fHandlers[i])->HandleData (fRxDMABuffer);
			}
		}
//		if (fRxDMABuffer->BufferCount () > 0 && lastAmount == fRxDMABuffer->BufferCount ()) {
//			((HCI *) fHandlers[0])->RemoveUnhandledPackets (fInputBuffer, fInputCurrent);
//		}
	} while (fRxDMABuffer->BufferCount () > 0 && lastAmount != fRxDMABuffer->BufferCount ());

	i = 0;
	while (i < fNumHandlers) {
		if (fHandlers[i]->fConnectionState == CONN_REMOVE) {
			RemoveHandler (fHandlers[i]);
		} else {
			i++;
		}
	}
}

void BluntServer::HandleCommand (BluntCommand* command)
{
	HCI* hci;
	int i;

	HLOG (1, "BluntServer::HandleCommand %d\n", command->fType);
	
	switch (command->fType) {
		case C_RESET: {
			BluntResetCommand* c = (BluntResetCommand*) command;
			HLOG (1, "  Reset: %s\n", c->fName);
			strcpy (((HCI *) fHandlers[0])->fName, c->fName);
			((HCI *) fHandlers[0])->Reset ();
			} break;
		case C_CONNECT: {
			BluntConnectionCommand* c = (BluntConnectionCommand*) command;
			HLOG (1, "  Connect: %d\n", c->fTargetLayer);
			SetupProtocolStack (c);
			} break;
		case C_DISCONNECT: {
			BluntDisconnectCommand* c = (BluntDisconnectCommand*) command;
			HLOG (1, "  Disconnect: %04x", c->fHCIHandle);
			for (i = 1; i < fNumHandlers; i++) {
				hci = (HCI *) fHandlers[i];
				if (hci->fHandlerType == H_HCI &&
					memcmp (c->fBdAddr, hci->fPeerAddress, 6) == 0) {
					hci->Disconnect (0x13);
					DeferredRemoveHandler (hci);
				}
			}
			} break;
		case C_INQUIRY: {
			BluntInquiryCommand* c = (BluntInquiryCommand*) command;
			HLOG (1, "  Inquiry %d %d\n", c->fTime, c->fAmount);
			((HCI *) fHandlers[0])->Inquiry (c->fTime, c->fAmount);
			} break;
		case C_INQUIRY_CANCEL: {
			HLOG (1, "  Cancel inquiry\n");
			((HCI *) fHandlers[0])->InquiryCancel ();
			} break;
		case C_NAME_REQUEST: {
			BluntNameRequestCommand* c = (BluntNameRequestCommand*) command;
			HLOG (1, "  Name request\n");
			((HCI *) fHandlers[0])->RemoteNameRequest (c->fBdAddr, c->fPSRepMode, c->fPSMode);
			} break;
		case C_INIT_PAIR:
			HLOG (1, "  Pair\n");
			InitiatePairing ((BluntInitiatePairingCommand*) command);
			break;
		case C_SERVICE_REQUEST:
			HLOG (1, "  Service request\n");
			InitiateServiceRequest ((BluntServiceRequestCommand*) command);
			break;
		case C_SET_LOG_LEVEL:
			HLOG (1, "  Set log level %d\n", ((BluntSetLogLevelCommand *) command)->fLevel);
			fLogLevel = ((BluntSetLogLevelCommand *) command)->fLevel;
			for (i = 0; i < fNumHandlers; i++) {
				fHandlers[i]->SetLogLevel (fLogLevel);
			}
			break;
		case C_DATA:
			HLOG (1, "  Send data\n");
			SendData ((BluntDataCommand*) command);
			break;
	}
	if (command->fDelete) delete command->fOriginalCommand;
}

void BluntServer::HandleTimer (BluntTimerEvent *event)
{
	Handler *handler = event->fHandler;
	HLOG (1, "BluntServer::HandleTimer %08x\n", handler);
	if (handler) {
		switch (handler->fHandlerType) {
			case H_HCI:
				((HCI *) handler)->HandleTimer (event->fUserData);
				break;
			case H_L2CAP:
				((L2CAP *) handler)->HandleTimer (event->fUserData);
				break;
			case H_SDP:
				((SDP *) handler)->HandleTimer (event->fUserData);
				break;
			case H_RFCOMM:
				((RFCOMM *) handler)->HandleTimer (event->fUserData);
				break;
		}
	} else {
		switch ((int) event->fUserData) {
			case EVT_BLUNT_CHECK_INPUT:
				break;
		}
	}
	delete event->fOriginalEvent;
}

void BluntServer::SetupProtocolStack (BluntConnectionCommand* command)
{
	HCI* hci;
	L2CAP* l2cap;
	RFCOMM* rfcomm;
	SDP* sdp;
	
	HLOG (1, "BluntServer::SetupProtocolStack\n");
	fNumHandlers = 1;

	hci = SetupHCILayer (command->fToolPort, command->fBdAddr, command->fPSRepMode, command->fPSMode);
	switch (command->fTargetLayer) {
		case L_HCI:
			hci->Connect ();
			break;
		case L_L2CAP:
			l2cap = SetupL2CAPLayer (command->fToolPort, command->fL2CAPProtocol, hci);
			l2cap->Connect ();
			break;
		case L_RFCOMM:
			l2cap = SetupL2CAPLayer (command->fToolPort, PSM_RFCOMM, hci);
			rfcomm = SetupRFCOMMLayer (command->fToolPort, command->fRFCOMMPort, l2cap);
			rfcomm->Connect ();
			break;
		case L_SDP:
			l2cap = SetupL2CAPLayer (command->fToolPort, PSM_SDP, hci);
			sdp = SetupSDPLayer (command->fToolPort, l2cap);
			sdp->Connect ();
			break;
	}
	
	Print ();
}

HCI *BluntServer::SetupHCILayer (TObjectId port, UChar* addr, UByte psRepMode, UByte psMode)
{
	HCI* hci;
	int i;

	HLOG (1, "BluntServer::SetupHCILayer\n");
	for (hci = NULL, i = 1; hci == NULL && i < fNumHandlers; i++) {
		if (memcmp (((HCI *) fHandlers[i])->fPeerAddress, addr, 6) == 0) hci = (HCI *) fHandlers[i];
	}
	if (hci == NULL) {
		hci = new HCI ();
		HLOG (1, "  HCI: %08x\n", hci);
		hci->fConnectionHandle = -1;
		hci->fTool = port;
		hci->fConnectionState = CONN_CONNECT;
		memcpy (hci->fPeerAddress, addr, 6);
		hci->fControlChannel = (HCI *) fHandlers[0];
		hci->fPSRepMode = psRepMode;
		hci->fPSMode = psMode;
		AddHandler (hci);
	}
	
	return hci;
}

L2CAP* BluntServer::SetupL2CAPLayer (TObjectId port, Short protocol, HCI* hci)
{
	L2CAP* controlL2cap;
	L2CAP* l2cap;
	int i;

	HLOG (1, "BluntServer::SetupL2CAPLayer\n");
	
	if (hci->fNumHandlers == 0) {
		controlL2cap = new L2CAP ();
		controlL2cap->fLocalCID = CID_SIGNAL;
		controlL2cap->fRemoteCID = CID_SIGNAL;
		hci->AddHandler ((Handler *) controlL2cap, this);
	} else {
		controlL2cap = (L2CAP *) hci->fHandlers[0];
	}
	controlL2cap->fTool = port;
	controlL2cap->fTargetProtocol = protocol;
	HLOG (1, " L2CAP control: %08x\n", controlL2cap);
	
	for (l2cap = NULL, i = 1; l2cap == NULL && i < hci->fNumHandlers; i++) {
		if (((L2CAP *) hci->fHandlers[i])->fProtocol == protocol) l2cap = (L2CAP *) hci->fHandlers[i];
	}
	if (l2cap == NULL) {
		l2cap = new L2CAP ();
		l2cap->fLocalCID = controlL2cap->GetNewCID ();
		l2cap->fProtocol = protocol;
		l2cap->fConnectionState = CONN_CONNECT;
		hci->AddHandler ((Handler *) l2cap, this);
	}
	l2cap->fTool = port;
	HLOG (1, " L2CAP data: %08x S: %d Prot: %d\n", l2cap, l2cap->fLocalCID, l2cap->fProtocol);
	
	return l2cap;
}

SDP* BluntServer::SetupSDPLayer (TObjectId port, L2CAP* l2cap)
{
	SDP* sdp;
	int i;
	
	HLOG (1, "BluntServer::SetupSDPLayer\n");
	for (sdp = NULL, i = 0; sdp == NULL && i < l2cap->fNumHandlers; i++) {
		if (l2cap->fHandlers[i]->fHandlerType == H_SDP) sdp = (SDP *) l2cap->fHandlers[i];
	}
	if (sdp == NULL) {
		sdp = new SDP ();
		sdp->fConnectionState = CONN_CONNECT;
		l2cap->AddHandler ((Handler *) sdp, this);
	}
	sdp->fTool = port;
	return sdp;
}

RFCOMM* BluntServer::SetupRFCOMMLayer (TObjectId port, UByte comPort, L2CAP* l2cap)
{
	RFCOMM* rfcomm;
	RFCOMM* controlRfcomm;
	
	HLOG (1, "BluntServer::SetupRFCOMMLayer\n");
	if (l2cap->fNumHandlers == 0) {
		controlRfcomm = new RFCOMM ();
		controlRfcomm->fTool = port;
		controlRfcomm->fDLCI = 0;
		controlRfcomm->fPort = 0;
		controlRfcomm->fConnectionState = CONN_CONNECT;
		l2cap->AddHandler ((Handler *) controlRfcomm, this);
	} else {
		controlRfcomm = (RFCOMM *) l2cap->fHandlers[0];
	}
	HLOG (1, " RFCOMM control: %08x\n", controlRfcomm);
	rfcomm = new RFCOMM ();
	rfcomm->fPort = comPort;
	rfcomm->fDLCI = comPort << 1;
	rfcomm->fTool = port;
	rfcomm->fConnectionState = CONN_CONNECT;
	l2cap->AddHandler ((Handler *) rfcomm, this);
	HLOG (1, " RFCOMM data: %08x Port: %d\n", rfcomm, rfcomm->fPort);

	return rfcomm;
}

Boolean BluntServer::RcvBufferLevelCritical ()
{
	if (fRxDMABuffer->BufferCount () > 2048) return true;
	else return false;
}

Boolean BluntServer::RcvBufferLevelOk (){
	if (fRxDMABuffer->BufferCount () < 256) return true;
	else return false;
}

void BluntServer::InitiatePairing (BluntInitiatePairingCommand* command)
{
	HCI* hci;

	HLOG (1, "BluntServer::InitiatePairing\n");
	
	hci = SetupHCILayer (0, command->fBdAddr, command->fPSRepMode, command->fPSMode);
	hci->Pair (command->fBdAddr, command->fPIN);
}

void BluntServer::InitiateServiceRequest (BluntServiceRequestCommand* command)
{
	HCI* hci;
	L2CAP* l2cap;
	SDP* sdp;

	HLOG (1, "BluntServer::InitiateServiceRequest\n");
	hci = SetupHCILayer (0, command->fBdAddr, command->fPSRepMode, command->fPSMode);
	l2cap = SetupL2CAPLayer (0, PSM_SDP, hci);
	sdp = SetupSDPLayer (0, l2cap);
	Print ();
	sdp->Connect ();
}

void BluntServer::SendData (BluntDataCommand* command)
{
	RFCOMM* rfcomm;
	Size count, n;
	int i;
	UByte data[192];
	Size mtu;
	BluntDataSentEvent *e;
	
	HLOG (1, "BluntServer::SendData %08x %d\n", command, command->fHandler->fHandlerType);
	if (command->fHandler->fHandlerType == H_RFCOMM) {
		rfcomm = (RFCOMM *) command->fHandler;
		mtu = MIN (192, rfcomm->fRemoteMTU - 16);
		count = command->fData->GetSize ();
		HLOG (1, "  MTU: %d, Len: %d\n", mtu, count);
		n = command->fData->Getn (data, mtu);
		if (n > 0) {
			rfcomm->SndUnnumberedInformation (data, n);
		} else {
			HLOG (1, "*** no bytes to send?\n");
		}
		TUPort p (rfcomm->fTool);
		e = new BluntDataSentEvent (noErr, n);
		p.Send ((TUAsyncMessage *) e, e, sizeof (BluntDataSentEvent), kNoTimeout, nil, BLUNT_MSG_TYPE);
	} else {
		TUPort p (rfcomm->fTool);
		e = new BluntDataSentEvent (kErrNoPeer, 0);
		p.Send ((TUAsyncMessage *) e, e, sizeof (BluntDataSentEvent), kNoTimeout, nil, BLUNT_MSG_TYPE);
	}
}

void BluntServer::BufferOutput (Boolean buffer)
{
	HLOG (2, "BluntServer::BufferOutput\n");
	fBufferOutput = buffer;
}

void BluntServer::SendEvent (BluntEvent* event)
{
	TUPort* port;
	
	HLOG (1, "BluntServer::SendEvent %08x\n", event);
	port = GetNewtTaskPort ();
	port->Send (event, (void *) event, event->GetSizeOf ());
}

void BluntServer::SetTimer (Handler *handler, int milliSecondDelay, void *userData)
{
	BluntTimerEvent* timer = new BluntTimerEvent (kResultOk, handler, userData);
	TTime delay;
	
	HLOG (1, "BluntServer::SetTimer %08x (%d) %d\n", handler, handler != NULL ? handler->fHandlerType : -1, milliSecondDelay);
	delay = GetGlobalTime () + TTime (milliSecondDelay, kMilliseconds);
	fPort.Send (timer, (void *) timer, sizeof (*timer), kNoTimeout, &delay, M_TIMER);
}

TObjectId BluntServer::Port ()
{
	TUNameServer nameServer;
	ULong id;
	ULong spec;

	nameServer.Lookup ("BluntServer", "TUPort", &id, &spec);
	return id;
}

void BluntServer::AddHandler (Handler *handler)
{
	HLOG (1, "BluntServer::AddHandler: %08x\n", handler);
	handler->fServer = this;
	fHandlers = (Handler**) realloc (fHandlers, (fNumHandlers + 1) *
		sizeof (Handler *));
	fHandlers[fNumHandlers] = handler;
	fNumHandlers++;
}

void BluntServer::RemoveHandler (Handler *handler)
{
	int i;
	
	HLOG (1, "BluntServer::RemoveHandler %08x\n", handler);
	for (i = 0; i < fNumHandlers; i++) {
		if (fHandlers[i] == handler) {
			memmove (&fHandlers[i], &fHandlers[i + 1],
				sizeof (Handler *) + fNumHandlers - 1);
			fNumHandlers--;
			if (handler->fHandlerType == H_HCI) {
				delete (HCI *) handler;
			} else {
				delete handler;
			}
			break;
		}
	}
	Print ();
}

void BluntServer::DeferredRemoveHandler (Handler *handler)
{
	int i;
	
	HLOG (1, "BluntServer::DeferredRemoveHandler %08x\n", handler);
	for (i = 0; i < fNumHandlers; i++) {
		if (fHandlers[i] == handler) {
			fHandlers[i]->fConnectionState = CONN_REMOVE;
			break;
		}
	}
}

void BluntServer::Print ()
{
	int i;
	
	HLOG (0, "BluntServer::Print\n");
	for (i = 0; i < fNumHandlers; i++) fHandlers[i]->Print (4);
}


void BluntServer::Log (int logLevel, char *format, ...)
{
    va_list args;
    char buffer[128];
    
    if (fLogLevel >= logLevel) {
        va_start (args, format);
        vsprintf (buffer, format, args);
        va_end (args);
        _sys_write (1, buffer, strlen (buffer));
    }
}

/*
void BluntServer::RxCAvailIntHandler (void)
{
	ULong s;
	
	while (fChip->RxBufFull () && fInputCurrent < MAX_INPUT) {
		s = fChip->GetRxErrorStatus ();
		if (s != 0) fRxStatus = s;
		fInputBuffer[fInputCurrent++] = fChip->GetByte ();
	}
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}
*/

/*
void BluntServer::HandleData ()
{
	int i;
	ULong lastAmount;
	
	HLOG (2, "BluntServer::HandleData %08x %d\n", fInputBuffer, fInputCurrent);
	lastAmount = 0;
	do {
		lastAmount = fInputCurrent;
		for (i = 0; fInputCurrent > 0 && i < fNumHandlers; i++) {
			HLOG (2, "BluntServer Handler: %d (%08x), data: %d bytes\n", i, fHandlers[i], fInputCurrent);
			((HCI *) fHandlers[i])->HandleData (fInputBuffer, fInputCurrent);
		}
	} while (lastAmount != fInputCurrent);

	if (fInputCurrent > 0) {
		do {
			lastAmount = fInputCurrent;
			((HCI *) fHandlers[0])->RemoveUnhandledPackets (fInputBuffer, fInputCurrent);
		} while (fInputCurrent > 0 && lastAmount != fInputCurrent);
	}
}
void BluntServer::InitiatePairing (BluntInitiatePairingCommand* command)
{
	HCI* hci;

	HLOG (1, "BluntServer::InitiatePairing\n");
	
	// set up the channel we're connecting over
	hci = new HCI ();
	hci->fConnectionHandle = -1;
	hci->fConnectionState = CONN_CONNECT;
	memcpy (hci->fPeerAddress, command->fBdAddr, 6);
	hci->fControlChannel = (HCI *) fHandlers[0];
	AddHandler (hci);
}
*/


/*
x:init(); x:start();
e := {_proto: @383}; e:Instantiate (e, {label: "rfcm", type: 'service, opCode: 512}); e:bind(nil, nil)
e:connect(nil, {async: true, reqtimeout: 20000, completionscript: func (a, b, c) begin end});
s := {form: 'binary, target: {data: MakeBinary (1000, 'bin)}, InputScript: func (e, data, term, opt) print ("input"), CompletionScript: func (e, opt, r) print (r)}
e:disconnect(nil, nil);
e:unbind(nil);
x:stop();

d := GetUnionSoup("Bluetooth Devices"):Query (nil):Entry()
o := {label: "addr", type: 'option, opCode: 512, data: {arglist: [[179, 136, 52, 55, 18, 0], 2], typelist: ['struct,	['array, 'byte, 6], 'byte]}}
o := {label: "addr", type: 'option, opCode: 512, data: {arglist: [[0, 96, 87, 2, 98, 166], 1], typelist: ['struct,	['array, 'byte, 6], 'byte]}}
o := {label: "addr", type: 'option, opCode: 512, data: {arglist: [[0, 16, 96, 41, 27, 145], 1], typelist: ['struct,	['array, 'byte, 6], 'byte]}}
f := {async: true, reqtimeout: 10000, completionscript: func (a, b, c) begin print (a && b && c) end}
c := {_proto: @383};
c:Instantiate (c, {label: "rfcm", type: 'service, opCode: 512});
c:bind(nil, nil)
c:connect(o, f)
c:disconnect(nil, nil)
c:cancel(nil)
c:unbind(nil)
c:dispose()

b := MakeBinary (3072, 'binary);
for i := 0 to length (b) - 1 do StuffByte (b, i, i mod 256);
getroot().|bluetooth setup:40hz|.fendpoint:output (b, nil, nil)

long BluntServer::TaskConstructor ()
{
	TUNameServer nameServer;
	TCMOSerialIOParms parms;
	SCCChannelInts handlers;
	NewtonErr r;
	ULong n;
	HCI* hci;
	TULockStack lock;
	
	HLOG (1, "BluntServer::TaskConstructor\n");
//	memset (fInputBuffer, 0, sizeof (fInputBuffer));
	fOutputCurrent = 0;
	fOutputCount = 0;
	fInputCurrent = 0;
	fInputHead = 0;
	fInputTail = 0;
    fDMA = true;

	fRxDMABuffer = new TExtendedCircleBuf ();
	fRxDMABuffer->Allocate (MAX_INPUT, 8, 2, 0);
	fTxDMABuffer = new TExtendedCircleBuf ();
	fTxDMABuffer->Allocate (MAX_OUTPUT, 8, 2, 0);
	fChip->SetSerialMode (kSerModeAsync);

	HLOG (1, "Server: %08x Debug: %08x Buffer: %08x\n", this, &fDebug, fRxDMABuffer);

	if (!fChip->PowerIsOn ()) fChip->PowerOn ();
	HLOG (1, "Power: %d\n", fChip->PowerIsOn ());
	DriverReset ();

	parms.fStopBits = k1StopBits;
	parms.fParity = kNoParity;
	parms.fDataBits = k8DataBits;
	parms.fSpeed = fSpeed;
	fChip->SetIOParms (&parms);

	fChip->SetInterruptEnable (false);
	LockStack (&lock, 64);
	EnterFIQAtomic ();
	fChip->Reconfigure ();
	if (fDMA) {
		fRxDMABuffer->DMABufInfo (&n, NULL, NULL, NULL);
		fTxDMABuffer->DMABufInfo (&n, NULL, NULL, NULL);
		fChip->InitRxDMA (fRxDMABuffer, MAX_INPUT, RxDmaIntHandler);
		fChip->InitTxDMA (fTxDMABuffer, TxDmaIntHandler);
		fChip->RxDMAControl (kDMANotifyOnNext | kDMAStart);
	}
	ExitFIQAtomic ();
	UnlockStack (&lock);

	HLOG (1, "Chip configured\n");
	
	handlers.TxBEmptyIntHandler = ::TxBEmptyIntHandler;
	handlers.ExtStsIntHandler = ::ExtStsIntHandler;
	handlers.RxCAvailIntHandler = ::RxCAvailIntHandler;
	handlers.RxCSpecialIntHandler = ::RxCSpecialIntHandler;
	
	LockHeapRange((VAddr) &::TxBEmptyIntHandler, (VAddr) &::__EndOfInterruptHandlers, true);
	r = fChip->InstallChipHandler (this, &handlers);
	HLOG (1, "InstallHandlers: %d\n", r);

//	Debugger();
//	HLOG (1, "Server code: %08x\n", *((int *) &BluntServer::TxBEmptyIntHandler));
	fChip->SetIntSourceEnable (kSerIntSrcTxBufEmpty, true);
	EnterFIQAtomic ();
	fChip->SetInterruptEnable (true);
	ExitFIQAtomic ();
	fChip->SetTxDTransceiverEnable (true);
	
	fNumHandlers = 0;
	fHandlers = NULL;
	hci = new HCI ();
	hci->fConnectionHandle = -1;
	AddHandler ((Handler *) hci);
	
	fPort.Init ();
	fIntMessage.Init (false);
	HLOG (1, "Server port: %08x\n", fPort.fId);
	nameServer.RegisterName ("BluntServer", "TUPort", fPort.fId, 0);
	
	fOutstandingPackets = 0;
	
	return noErr;
}

*/
